"""This file contains tests for creating overflow graph. ie,
graph with delta flows, current_flows - flows_after_closing_line"""

import numpy as np
import configparser
import networkx as nx
from alphaDeesp.core.network import Network
from alphaDeesp.core.grid2op.Grid2opObservationLoader import Grid2opObservationLoader
from alphaDeesp.core.grid2op.Grid2opSimulation import Grid2opSimulation#, build_powerflow_graph
from alphaDeesp.core.graphsAndPaths import OverFlowGraph, PowerFlowGraph
from alphaDeesp.core.alphadeesp import AlphaDeesp



def build_sim():
    config = configparser.ConfigParser()
    config.read("./alphaDeesp/tests/resources_for_tests_grid2op/config_for_tests.ini")
    param_folder = "./alphaDeesp/tests/resources_for_tests_grid2op/l2rpn_2019_ltc_9"

    loader = Grid2opObservationLoader(param_folder)
    env, obs, action_space = loader.get_observation()
    observation_space = env.observation_space
    sim = Grid2opSimulation(obs, action_space, observation_space, param_options=config["DEFAULT"], debug=False,
                                 ltc=[9])
    return sim,env


def test_powerflow_graph():
    sim,env = build_sim()

    g_pow = PowerFlowGraph(sim.topo, sim.lines_outaged).get_graph()
    print(sim.df)
    path_to_saved_graph = "./alphaDeesp/tests/resources_for_tests_grid2op/saved_graphs/g_pow.dot"
    saved_g = nx.drawing.nx_pydot.read_dot(path_to_saved_graph)
    
    for e1 in saved_g.edges(data="label"):
        for e2 in g_pow.edges(data="label"):
            if int(e1[0]) == int(e2[0]) and int(e1[1]) == int(e2[1]):
                saved_flow = float(e1[2])
                current_flow = float(e2[2])
                assert (saved_flow == current_flow)

    print("g_over and saved_g are isomorphic: ", nx.is_isomorphic(g_pow, saved_g))
    #assert (nx.is_isomorphic(g_pow, saved_g))


def test_overflow_grid():
    """This function, given the input folder in test/path,
    it computes the overflow graph and compares it with the saved graph: saved_overflow_graph.dot"""
    sim,env = build_sim()
    df_of_g = sim.get_dataframe()
    g_over = OverFlowGraph(sim.topo, [9], df_of_g).get_graph()
    path_to_saved_graph = "./alphaDeesp/tests/resources_for_tests_grid2op/saved_graphs/g_over.dot"

    saved_g = nx.drawing.nx_pydot.read_dot(path_to_saved_graph)

    for e1 in saved_g.edges(data="label"):
        for e2 in g_over.edges(data="label"):
            if int(e1[0]) == int(e2[0]) and int(e1[1]) == int(e2[1]):
                saved_flow = float(e1[2])
                current_flow = float(e2[2])
                assert (saved_flow == current_flow)

    print("g_over and saved_g are isomorphic: ", nx.is_isomorphic(g_over, saved_g))


def test_new_flows_after_network_cut():
    """After closing one electric line, the flows should change on the network."""
    sim,env = build_sim()

    new_flows = sim.cut_lines_and_recomputes_flows([9])

    print("diff flows = ")
    print((new_flows - sim.obs.p_or).astype(int))
    result = (new_flows - sim.obs.p_or).astype(int)
    print(result != 0.0)
    print("there are {} non zero elements".format(np.count_nonzero(result)))

    # only the line that got cut should be at zero.
    assert (np.count_nonzero(result) == 19)


def extract_topology():
    """This function takes an observation and tests the function: extract_topo_from_obs"""
    print("Test extract_topology")
    sim,env = build_sim()
    print("flows = ", sim.obs.p_or.astype(int))


def test_isgraph_connected():
    """This function test if a graph is connexe"""
    # Three different types of graph for tests
    # ============ Graph g1 ============
    #        5
    #         \>
    # 1->-2->-3->-4
    #      \>
    #      6
    # printer = Printer()

    g1 = nx.DiGraph()
    g1.add_edges_from([(1, 2), (2, 3), (2, 6), (3, 4), (5, 3)])
    # printer.display_geo(g1, name="test_graph_connected")
    print("graph g1 is connected: ", nx.is_weakly_connected(g1))
    assert (nx.is_weakly_connected(g1))
    g1.remove_edge(2, 3)
    print("After removing edge graph g1 is not_connected: ", nx.is_weakly_connected(g1))
    assert (not nx.is_weakly_connected(g1))


def test_detailed_graph_with_node_0_split_in_two():
    """
    From an observation, avec using load2()
    basic graph with zero nodes split in two, has 14 nodes
    with node 0 split in two, we should have 15 nodes
    :return:
    """
    sim,env = build_sim()
    new_configuration = [[2, 2, 2, 1, 1]]
    node_id = [4]
    _current_observation = sim.change_nodes_configurations(new_configuration, node_id, env)
    print(_current_observation)

    network = Network(sim.substations_elements)
    print("There are {} graphical nodes ".format(network.get_graphical_number_of_nodes()))
    assert (network.get_graphical_number_of_nodes() == 15)

def test_apply_topo():
    """
    Load configuration, modifiy topo on substation 4 with grid2op API, and generate graphs
    Then apply new topo with alphadeesp core function
    :return:
    """
    # import os
    # os.chdir('../../../')

    config = configparser.ConfigParser()
    config.read("./alphaDeesp/tests/resources_for_tests_grid2op/config_for_tests.ini")
    param_folder = "./alphaDeesp/tests/resources_for_tests_grid2op/l2rpn_2019_ltc_9"
    ltc = [9]

    loader = Grid2opObservationLoader(param_folder)
    env, obs, action_space = loader.get_observation()

    sub_id = 4
    new_configuration = [2, 2, 2, 1, 1] # Grid2op notation
    new_conf_alphadeesp_1 = [1, 1, 0, 0, 0] # ALphadeesp notation
    new_conf_alphadeesp_2 = [0, 0, 0, 0, 0] # ALphadeesp notation

    action_def = {"set_bus": {"substations_id": [(sub_id, new_configuration)]}}
    action = env.action_space(action_def)
    osb, _reward, _done, _info = env.step(action)

    observation_space = env.observation_space
    sim = Grid2opSimulation(obs, action_space,observation_space, param_options=config["DEFAULT"], debug=False,
                            ltc=ltc)

    # Get data from simulator representing the grid before and after line cutting, and topologies. Then initialize ALphadeesp
    df_of_g = sim.get_dataframe()
    g_over = OverFlowGraph(sim.topo, ltc, df_of_g).get_graph()
    #g_pow = sim.build_powerflow_graph_beforecut()
    #g_pow_prime = sim.build_powerflow_graph_aftercut()
    simulator_data = {"substations_elements": sim.get_substation_elements(),
                      "substation_to_node_mapping": sim.get_substation_to_node_mapping(),
                      "internal_to_external_mapping": sim.get_internal_to_external_mapping()}
    printer = None
    custom_layout = sim.get_layout()
    debug = False
    g_copy = g_over.copy() # Very important step ! otherwise, apply_new_topo will change g_over and it won't be chainable
    alphadeesp = AlphaDeesp(g_over, df_of_g,simulator_data, sim.substation_in_cooldown, debug=debug)


    # Apply topo twice with Alphadeesp core on this Network and check if applied correctly
    new_graph, internal_repr = alphadeesp.apply_new_topo_to_graph(g_copy, new_topology=new_conf_alphadeesp_1,
                                                                  node_to_change=sub_id)
    for i, busbar_id in enumerate(new_conf_alphadeesp_1):
        assert (internal_repr[sub_id][i].busbar_id == busbar_id)

    new_graph, internal_repr = alphadeesp.apply_new_topo_to_graph(g_copy, new_topology=new_conf_alphadeesp_2,
                                                                  node_to_change=sub_id)
    for i, busbar_id in enumerate(new_conf_alphadeesp_2):
        assert (internal_repr[sub_id][i].busbar_id == busbar_id)






# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of ExpertOp4Grid, an expert system approach to solve flow congestions in power grids

""" This file is the main file for the Expert Agent called AlphaDeesp """
import networkx as nx
import pandas as pd
import itertools
import pprint
import numpy as np

from alphaDeesp.core.graphsAndPaths import Structured_Overload_Distribution_Graph
from alphaDeesp.core.elements import *
from math import fabs, ceil


import os


# os.environ['PATH'] += os.pathsep + r'C:\Users\nmegel\graphviz-2.38\release\bin'


class AlphaDeesp:  # AKA SOLVER
    def __init__(self, _g, df_of_g, simulator_data=None, substation_in_cooldown=[], debug=False):
        # used for postprocessing
        self.bag_of_graphs = {}
        self.debug = debug
        self.boolean_dump_data_to_file = False

        # data from Simulator Class
        self.simulator_data = simulator_data

        self.data = {}
        self.g = _g  # here the g is the overflow graph
        self.df = df_of_g
        self.initial_graph = self.g.copy()
        self.substation_in_cooldown = substation_in_cooldown  # we cannot play with those substations so no need to compute simulations

        # check that line extemity does not have only load or productions: otherwise there is either node merging to do or nothing else

        ranked_combinations_structure_initiation = {
            "score": ["XX"],
            "topology": [["X", "X", "X"]],
            "node": ["X"]
        }
        ranked_combinations = pd.DataFrame(ranked_combinations_structure_initiation)

        #Compute the overload distribution graph (constrained path, loops, hubs)
        self.g_distribution_graph=Structured_Overload_Distribution_Graph(self.g)

        # this function takes the dataFrame self.red_loops and adds the min cut_values to it.
        self.rank_red_loops()

        self.rankedLoopBuses = self.rank_loop_buses(self.g, self.df)

        # here we classify nodes into 4 categories
        self.structured_topological_actions = self.identify_routing_buses()  # it is a dict
        # print("#########################################################################")
        # print("structured_top_actions =", self.structured_topological_actions)
        # print("#########################################################################")

        self.ranked_combinations = self.compute_best_topologies()

        # for ranked_comb in self.ranked_combinations:
        # print("---------------------------")
        # print(ranked_comb)

        if self.boolean_dump_data_to_file:
            self.ranked_combinations[0].to_csv("./result_ranked_combinations.csv", index=True)

        # The order to get structures is:
        # constrained_path
        # hubs
        # parallel loops and loops

        # self.simulate_network_change(self.ranked_combinations)

    def load2(self, observation, line_to_cut: int):
        """@:arg observation: a pypownet observation,
        line_to_cut: line to cut for overload graph"""

    def simulate_network_change(self, ranked_combinations):
        """This function takes a dataFrame ranked_combinations and computes new changes with Pypownet"""
        pass

    def get_ranked_combinations(self):
        return self.ranked_combinations

    def compute_best_topologies(self):
        """
        inputs: selected_ranked_nodes (after having identified routing buses, ie 4 categories)
        :return: pd.DataFrame containing all topologies scored.
        """
        selected_ranked_nodes = []
        for key_indice in list(self.structured_topological_actions.keys()):
            res = self.structured_topological_actions[key_indice]
            if res is not None:
                for elem in res:
                    selected_ranked_nodes.append(elem)

        res_container = []
        for node in selected_ranked_nodes:
            if node in self.substation_in_cooldown:
                print("substation " + str(node) + " is in cooldown and no action can be performed on it for now")
                continue

            all_combinations = self.compute_all_combinations(node)
            if (len(all_combinations) != 0):
                ranked_combinations = self.rank_topologies(all_combinations, self.g, node)
                # print(ranked_combinations)

                # best_topologies = best_topologies.append(ranked_combinations)
                # pd.concat([best_topologies, *ranked_combinations])

                # print("\n##############################################################################")
                # print("##########............BEST_TOPOLOGIES COMPUTED............####################")
                # print("##############################################################################")

                # best_topologies = self.clean_and_sort_best_topologies(best_topologies)
                best_topologies = self.clean_and_sort_best_topologies(ranked_combinations)
                res_container.append(best_topologies)
            # # print(best_topologies)

        return res_container

    def clean_and_sort_best_topologies(self, best_topologies):
        """This function cleans the Dataframe best_topologies;
        it deletes rows with XX, and sorts the Dataframe. In order to achieve this we have to set_index first."""
        best_topologies = best_topologies.set_index("score")
        best_topologies = best_topologies.drop("XX", axis=0)
        best_topologies = best_topologies.sort_values("score", ascending=False)

        return best_topologies

    def compute_all_combinations(self, node):
        """ Given a node, returns all possible combinations of a node configuration.
        ex: [001], [010], [100], [101], [011]... etc..."""

        # ## check that current topology is not in this list
        # TO DO: manage the fact that a substation can already be in 2 nodes. How do you get the node configuration and everything?
        node_configuration_elements = self.simulator_data["substations_elements"][node]
        n_elements = len(node_configuration_elements)
        node_configuration = [node_configuration_elements[i].busbar_id for i in range(n_elements)]
        node_configuration_sym = [0 if node_configuration[i] == 1 else 1 for i in range(n_elements)]

        # print("Inside compute_all_comb : for node [{}], node_configuration = {}".format(node, node_configuration))
        if n_elements == 0 or n_elements == 1:
            raise ValueError("Cannot generate combinations out of a configuration with len = 1 or 2")
        elif n_elements == 2:
            return [(1, 1), (0, 0)]
        else:
            l = [0, 1]
            allcomb = [list(i) for i in itertools.product(l, repeat=n_elements)]

            #we also want to filter combs that only have prods and loads connected to a node
            nProds_loads=0
            for element in node_configuration_elements:
                if isinstance(element, Production) or isinstance(element, Consumption):
                    nProds_loads+=1
                else:
                    break

            # we get rid of symetrical topologies by fixing the first element to busbar 0.
            # ideally if first element is not connected, we should fix the first connected element
            # Also a node should also have 2 elements connected to it, we filter that as well
            uniqueComb = [allcomb[i] for i in range(len(allcomb)) if self.legal_comb(allcomb[i],nProds_loads,n_elements,node_configuration,node_configuration_sym)]
                          #if (allcomb[i][0] == 0) & (allcomb[i] != node_configuration) & (allcomb[i] != node_configuration_sym) &
                          #(np.sum(allcomb[i]) != 1) & (np.sum(allcomb[i]) != n_elements - 1) &
                          #]  # we get rid of symetrical topologies by fixing the first element to busbar 0. ideally if first element is not connected, we should fix the first connected element


        return uniqueComb

    def legal_comb(self,comb,nProd_loads,n_elements,node_configuration,node_configuration_sym):
        sum_comb=np.sum(comb)
        busBar_prods_loads=set(comb[0:nProd_loads])
        busBar_lines = set(comb[nProd_loads:])

        areProdsLoadsIsolated=False
        if(nProd_loads>=2) and (sum_comb != 1) and (sum_comb != n_elements - 1):
            busbar_diff=set(busBar_prods_loads)-set(busBar_lines)
            if(len(busbar_diff)!=0):
                areProdsLoadsIsolated=True

        legal_condition=((comb[0] == 0) & (comb != node_configuration) & (comb != node_configuration_sym) &
        (sum_comb != 1) & (sum_comb != n_elements - 1) & (areProdsLoadsIsolated==False))

        return legal_condition

    def rank_topologies(self, all_combinations, graph, node_to_change: int):
        """==> ultimate goal: This function returns a DF with topologies ranked
        for the moment:
            takes a topo,
            apply it to graph,
            compute score,
            add to df
            next
        """

        ranked_combinations_columns = ["score", "topology", "node"]
        scores_data = [["XX", ["X", "X", "X"], "X"]]

        # ===========================================
        # print("\nNOEUD "+str(node_to_change))
        # print("number of topo to test "+str(len(all_combinations)))

        for i, topo in enumerate(all_combinations):
            # WARNING the internal_repr is not used further in the code. It is not up to date with the new_graph.
            # Only the original one.
            isSingleNodeTopo = ((np.all(np.array(topo) == 0)) or (np.all(np.array(topo) == 1)))
            score = self.rank_current_topo_at_node_x(self.g, node_to_change, isSingleNodeTopo, topo)
            if self.debug:
                print("\n** RESULTS ** new topo [{}] on node [{}] has a score: [{}]\n".format(topo, node_to_change, score))
            scores_data.append([score, topo, node_to_change])

            # =======================================================
            # if i % 10000 ==0:
              #   print("Done: "+str(i)+" topos")

        ranked_combinations = pd.DataFrame(columns = ranked_combinations_columns, data = scores_data)

        # =================================================
        # ranked_combinations.to_csv("NEW_rank_topologies_l2rpn_2019_node_"+str(node_to_change)+".csv", sep = ';', decimal = ',')
        return ranked_combinations

    # WARNING: does not work yet when you go back from two nodes to one node at a given substation? Basically one node will be not connected?
    def apply_new_topo_to_graph(self, graph: nx.MultiDiGraph, new_topology, node_to_change: int):
        """given  a graph, a node_topoly and a node_id, this function applies the change to the graph
        :return new_graph, internal_repr_dict"""
        if self.debug:
            print("\n====================================== apply new topo to graph ======================================")
            print(" new topology applied = [{}] to node: [{}]".format(new_topology, node_to_change))
            print("======================================================================================================\n")

        # check if there are two nodes, there are 2 different values in new topo 0 and 1
        bus_ids = set(new_topology)
        # assert(len(bus_ids) == 2)#not necesarrily, it should be at least 1 and not more than 2
        assert ((len(bus_ids) != 0) & (len(bus_ids) <= 2))

        internal_repr_dict = dict(self.simulator_data["substations_elements"])
        new_node_id = int("666" + str(node_to_change))

        element_types = self.simulator_data["substations_elements"][node_to_change]
        # TO DO: you need to get all elements of the substation, especially if it is already in a 2 nodes topology and you want to merge the nodes
        # if (new_node_id in self.simulator_data["substations_elements"]):#the substation was already in a 2 node topology
        #    element_types+=self.simulator_data["substations_elements"][new_node_id]
        # it has to be the same, otherwise it does not make sense, ie, there is an error somewhere
        assert len(element_types) == len(new_topology)

        # BEFORE REMOVING, GET NEEDED INFORMATION ON EDGES: COLORS, WIDTH etc...
        color_edges = {}
        for u, v,idx, color in self.g.edges(data="color",keys=True):
            # invert edges that have been marked as SWAPPED in DATAFRAME.
            condition = list(self.df.query("idx_or == " + str(u) + " & idx_ex == " + str(v))["swapped"])[0]
            color_edges[(u, v,idx)] = color
            if condition:
                color_edges[(v, u,idx)] = color
            else:
                color_edges[(u, v,idx)] = color

        if 1 in new_topology:  # ie, if the topo is not [0, ... , 0]
            # first we delete the node_to_change ==> it deletes all edges for us
            graph.remove_node(node_to_change)

        # ################ PREPROCESSING NODE RECONSTRUCTION PART, IMPORTANT TO GET COLORS RIGHT
        i = 0
        current_node = []  # Busbar 0
        new_node = []  # Busbar 1
        # then, parsing element by element, reconnect the graph.
        for internal_elem, element, element_type in zip(internal_repr_dict[node_to_change], new_topology, element_types):
            internal_elem.busbar_id = element

        # WE RECONSTRUCT INTERNAL REPR
        # prod = {0: 0, 1: 0}  # busid:value
        # load = {0: 0, 1: 0}  # busid:value
        prod = {}
        load = {}
        for element in internal_repr_dict[node_to_change]:
            if element.busbar_id not in prod.keys():
                if isinstance(element, Production):
                    prod[element.busbar_id] = fabs(element.value)
            else:
                if isinstance(element, Production):
                    prod[element.busbar_id] += fabs(element.value)

            if element.busbar_id not in load.keys():
                if isinstance(element, Consumption):
                    load[element.busbar_id] = fabs(element.value)
            else:
                if isinstance(element, Consumption):
                    load[element.busbar_id] += fabs(element.value)

        node_type = {}
        for bus_id in bus_ids:  # [0, 1]
            node_label = node_to_change
            prod_minus_load = 0

            if bus_id in prod.keys() and bus_id in load.keys():
                prod_minus_load = prod[bus_id] - load[bus_id]
                if prod_minus_load > 0:
                    node_type[bus_id] = "prod"
                else:
                    node_type[bus_id] = "load"
            elif bus_id in prod.keys():
                node_type[bus_id] = "prod"
                prod_minus_load = prod[bus_id]
            elif bus_id in load.keys():
                node_type[bus_id] = "load"
                prod_minus_load = load[bus_id]

            if bus_id == 1:
                node_label = new_node_id
            if bus_id in node_type.keys():
                if node_type[bus_id] == "prod":  # PROD
                    # prod_minus_load = prod[bus_id]
                    graph.add_node(node_label, pin=True, prod_or_load="prod", value=str(prod_minus_load),
                                   style="filled", fillcolor="#f30000")
                elif node_type[bus_id] == "load":
                    # prod_minus_load = load[bus_id]
                    graph.add_node(node_label, pin=True, prod_or_load="load", value=str(-prod_minus_load),
                                   style="filled", fillcolor="#478fd0")
            else:  # WHITE
                graph.add_node(node_label, pin=True, prod_or_load="load", value=str(prod_minus_load),
                               style="filled", fillcolor="#ffffff")

        i = 0
        # then, parsing element by element, reconnect the graph.
        for element, element_type in zip(new_topology, element_types):
            # print("element = ", element)
            # print("element type = ", element_type)
            reported_flow = None
            edge_color = None
            penwidth = None
            if isinstance(element_type, OriginLine) or isinstance(element_type, ExtremityLine):
                # # print("element_type.flow_value=", element_type.flow_value)
                reported_flow = element_type.flow_value[0]
                # # print("reported_flow = ", reported_flow)
                penwidth = fabs(reported_flow) / 10
                if penwidth == 0.0:
                    penwidth = 0.1

            if element == 1:  # connect FROM or TO node: 666XX
                if isinstance(element_type, OriginLine):
                    graph.add_edge(new_node_id, element_type.end_substation_id,
                                   capacity=float("%.2f" % reported_flow), label="%.2f" % reported_flow,
                                   color=color_edges[(node_to_change, element_type.end_substation_id,0)],
                                   fontsize=10, penwidth="%.2f" % penwidth)#we have a multiGraph, so we need to give a third index to color_edges, should be revised if reused

                elif isinstance(element_type, ExtremityLine):
                    graph.add_edge(element_type.start_substation_id, new_node_id,
                                   capacity=float("%.2f" % reported_flow), label="%.2f" % reported_flow,
                                   color=color_edges[(element_type.start_substation_id, node_to_change,0)],
                                   fontsize=10, penwidth="%.2f" % penwidth)

            elif element == 0:  # connect from node node:_to_change
                if isinstance(element_type, OriginLine):
                    graph.add_edge(node_to_change, element_type.end_substation_id,
                                   capacity=float("%.2f" % reported_flow), label="%.2f" % reported_flow,
                                   color=color_edges[(node_to_change, element_type.end_substation_id,0)],
                                   fontsize=10, penwidth="%.2f" % penwidth)

                elif isinstance(element_type, ExtremityLine):
                    graph.add_edge(element_type.start_substation_id, node_to_change,
                                   capacity=float("%.2f" % reported_flow), label="%.2f" % reported_flow,
                                   color=color_edges[(element_type.start_substation_id, node_to_change,0)],
                                   fontsize=10, penwidth="%.2f" % penwidth)

            else:
                raise ValueError("Error element has to belong to either Busbar 1 or 2")

            i += 1

        name = "".join(str(e) for e in new_topology)
        name = str(node_to_change) + "_" + name
        self.bag_of_graphs[name] = graph
        return graph, internal_repr_dict

    def rank_current_topo_at_node_x(self, graph, node: int, isSingleNode=False, topo_vect=[0, 0, 1, 1, 1],is_score_specific_substation=True):
        """This function ranks current topology at node X"""
        final_score = 0.0
        all_nodes_value_attributes = nx.get_node_attributes(graph, "value")  # dict[node]
        all_edges_color_attributes = nx.get_edge_attributes(graph, "color")  # dict[edge]
        all_edges_xlabel_attributes = nx.get_edge_attributes(graph, "label")  # dict[edge]

        # ======================================
        # print('\nnoeud '+str(node)+' topo '+str(topo_vect))

        #  ########## IS IN AMONT ##########
        constrained_path=self.g_distribution_graph.get_constrained_path()
        red_loops = self.g_distribution_graph.get_loops()

        in_negative_flows = []
        out_negative_flows = []
        out_positive_flows = []
        in_positive_flows = []

        if node in constrained_path.n_amont():
            # ======================================
            #print("AMONT")
            if self.debug:
                print("||||||||||||||||||||||||||| node [{}] is in_Amont of constrained_edge".format(node))

            interesting_bus_id = 0
            if is_score_specific_substation:#when comparing all combinations at a substation, you don't need to figure out for which of the two splitted nodes you compute the score, as the two symetrical combinations on the 2 nodes exist
                for edge in graph.out_edges(node,keys=True):
                    if self.is_connected_to_cpath(all_edges_color_attributes, all_edges_xlabel_attributes, node, edge, isSingleNode):
                        # take the other bus id
                        interesting_bus_id = abs(self.get_bus_id_from_edge(node, edge, topo_vect) - 1)
                        break
            else:#when comparing all combinations at a substation, you don't need to figure out for which of the two splitted nodes you compute the score, as the two symetrical combinations on the 2 nodes exist
                #find most interesting node with largest ingoing negative flow
                in_edge_capacities_bus0 = [float(all_edges_xlabel_attributes[edge]) for edge in graph.in_edges(node, keys=True)
                                      if self.get_bus_id_from_edge(node, edge, topo_vect) == 0]
                in_edge_negative_capacities_bus0 = fabs(sum(x for x in in_edge_capacities_bus0 if x < 0))

                in_edge_capacities_bus1 = [float(all_edges_xlabel_attributes[edge]) for edge in
                                           graph.in_edges(node, keys=True)
                                           if self.get_bus_id_from_edge(node, edge, topo_vect) == 1]
                in_edge_negative_capacities_bus1 = fabs(sum(x for x in in_edge_capacities_bus1 if x < 0))

                if in_edge_negative_capacities_bus1>in_edge_negative_capacities_bus0:
                    interesting_bus_id=1

            not_interesting_bus_id=fabs(interesting_bus_id-1)

            # somme des reports négatifs entrants + sommes des reports positifs entrants
            for edge in graph.in_edges(node,keys=True):
                edge_bus_id = self.get_bus_id_from_edge(node, edge, topo_vect)
                if edge_bus_id == interesting_bus_id: # MASK
                    edge_flow_value = float(all_edges_xlabel_attributes[edge])
                    if edge_flow_value < 0:
                        in_negative_flows.append(fabs(edge_flow_value))
                    else:
                        in_positive_flows.append(edge_flow_value)

            # somme des reports positifs sortant +
            for edge in graph.out_edges(node,keys=True):
                edge_bus_id = self.get_bus_id_from_edge(node, edge, topo_vect)
                if edge_bus_id == interesting_bus_id: # MASK
                    edge_flow_value = float(all_edges_xlabel_attributes[edge])
                    if edge_flow_value > 0:
                        out_positive_flows.append(edge_flow_value)
                    else:
                        out_negative_flows.append(fabs(edge_flow_value))

            diff_sums = self.get_prod_conso_sum(node, interesting_bus_id, topo_vect)#-self.get_prod_conso_sum(node, not_interesting_bus_id, topo_vect) #it is better to make the balance between the two nodes, did we put more generation on this one than the other one
            max_pos_in_or_out_flows = max(sum(out_positive_flows), sum(in_positive_flows))#min(sum(out_positive_flows), sum(in_positive_flows))#
            #we want to push ingoing negative  and production towards red path,
            #final_score = np.around(sum(in_negative_flows) - np.around(sum(out_negative_flows)) + max_pos_in_or_out_flows + diff_sums, decimals=2)#+ diff_sums, it is not very clear its impact
            if (is_score_specific_substation):
                final_score = np.around(
                    sum(in_negative_flows) + max_pos_in_or_out_flows + diff_sums,
                    decimals=2)
            else:
                final_score = np.around(
                    sum(in_negative_flows) - np.around(sum(out_negative_flows)) + sum(out_positive_flows),# + diff_sums,
                    decimals=2)  # + diff_sums, it is not very clear its impact
            if self.debug:
                print("AMONT")
                print("diff_sums = ", diff_sums)
                print(type(diff_sums))
                print("max_pos_in_or_out_flows = ", max_pos_in_or_out_flows)
                print("in negative flows = ", in_negative_flows)
                print("out negative flows = ", out_negative_flows)
                print("out positive flow = ", out_positive_flows)
                print("Final score = ", final_score)

        #  ########## IS IN AVAL ##########
        elif node in constrained_path.n_aval():
            # =============================================================
            #print("AVAL")
            if self.debug:
                print("||||||||||||||||||||||||||| node [{}] is in_Aval of constrained_edge".format(node))

            interesting_bus_id = 0

            if is_score_specific_substation:#when comparing all combinations at a substation, you don't need to figure out for which of the two splitted nodes you compute the score, as the two symetrical combinations on the 2 nodes exist
                for edge in graph.in_edges(node,keys=True):
                    if self.is_connected_to_cpath(all_edges_color_attributes, all_edges_xlabel_attributes, node, edge, isSingleNode):
                        # take the other bus id
                        interesting_bus_id = abs(self.get_bus_id_from_edge(node, edge, topo_vect) - 1)
                        break
            else: #when you are considering only specific configurations (without its symetric one) and want to compare this among substations, not only within this substation
                # find most interesting node with largest ougoing negative flow
                out_edge_capacities_bus0 = [float(all_edges_xlabel_attributes[edge]) for edge in
                                            graph.out_edges(node, keys=True)
                                            if self.get_bus_id_from_edge(node, edge, topo_vect) == 0]
                out_edge_negative_capacities_bus0 = fabs(sum(x for x in out_edge_capacities_bus0 if x < 0))

                out_edge_capacities_bus1 = [float(all_edges_xlabel_attributes[edge]) for edge in
                                            graph.out_edges(node, keys=True)
                                            if self.get_bus_id_from_edge(node, edge, topo_vect) == 1]
                out_edge_negative_capacities_bus1 = fabs(sum(x for x in out_edge_capacities_bus1 if x < 0))

                if out_edge_negative_capacities_bus1 > out_edge_negative_capacities_bus0:
                    interesting_bus_id = 1

            # somme des reports negatifs et positifs SORTANT
            for edge in graph.out_edges(node,keys=True):
                edge_bus_id = self.get_bus_id_from_edge(node, edge, topo_vect)
                if edge_bus_id == interesting_bus_id: # MASK
                    edge_flow_value = float(all_edges_xlabel_attributes[edge])
                    if edge_flow_value < 0:
                        out_negative_flows.append(fabs(edge_flow_value))
                    else:
                        out_positive_flows.append(edge_flow_value)

            for edge in graph.in_edges(node,keys=True):
                edge_bus_id = self.get_bus_id_from_edge(node, edge, topo_vect)
                if edge_bus_id == interesting_bus_id: # MASK
                    edge_flow_value = float(all_edges_xlabel_attributes[edge])
                    if edge_flow_value > 0:
                        in_positive_flows.append(edge_flow_value)
                    else:
                        in_negative_flows.append(fabs(edge_flow_value))

            # MAX (somme des reports positifs ENTRANT ET SORTANT)
            max_pos_in_or_out_flows = max(sum(out_positive_flows), sum(in_positive_flows))

            if self.debug:
                print("out_negative_flows = ", out_negative_flows)
                print("in_negative_flows = ", in_negative_flows)
                print("out_positive_flows = ", out_positive_flows)
                print("in_positive_flows = ", in_positive_flows)
                print("sum out neg = ", sum(out_negative_flows))
                print("sum out neg = ", sum(in_negative_flows))
                print("sum out pos = ", sum(out_positive_flows))
                print("sum in  pos = ", sum(in_positive_flows))
                print("max_pos_in_or_out_flows = ", max_pos_in_or_out_flows)
            # sur le noeud choisi (non connecté au cpath) on souhaite y connecter des consommations et pas des
            # productions pour l'aval.
            diff_sums = -(self.get_prod_conso_sum(node, interesting_bus_id, topo_vect)) #- self.get_prod_conso_sum(node,not_interesting_bus_id,topo_vect))
            if (is_score_specific_substation):
                final_score = np.around(sum(out_negative_flows)+ max_pos_in_or_out_flows + diff_sums, decimals=2)
            else:
                final_score = np.around(
                sum(out_negative_flows) - np.around(sum(in_negative_flows)) + sum(in_positive_flows),# + diff_sums,
                decimals=2)

            if self.debug:
                print("AVAL")
                print("diff_sums = ", diff_sums)
                print("Final score = ", final_score)
                print(type(final_score))

        #  ########## IS IN Loop ##########
        # you want a node with the maximum output lines connected to the ingoing red loop edges, not connected to other ingoing edges
        elif node in set([x for loop in range(len(red_loops.Path)) for x in red_loops.Path[loop]]):
            # ========================================================================
            # print("AUTRE")
            node2 = int("666" + str(node))

            if 1 in topo_vect and 0 in topo_vect:  # need to be a 2 node topology
                # we find the node with the biggest red ingoing delta flow
                InputRedDeltaFlow_1 = 0
                InputRedDeltaFlow_2 = 0


                for edge in graph.in_edges(node,keys=True):
                    edge_bus_id = self.get_bus_id_from_edge(node, edge, topo_vect)
                    edge_color = all_edges_color_attributes[edge]
                    edge_value = float(all_edges_xlabel_attributes[edge])
                    if (edge_color == "coral"):
                        if edge_bus_id == 0: # MASK
                            InputRedDeltaFlow_1 += edge_value
                        elif edge_bus_id == 1: # MASK
                            InputRedDeltaFlow_2 += edge_value

                Bus_BiggestInputDeltaFlow = 0
                InputRedDeltaFlow = InputRedDeltaFlow_1
                if (InputRedDeltaFlow_2 >= InputRedDeltaFlow_1):
                    Bus_BiggestInputDeltaFlow = 1
                    InputRedDeltaFlow = InputRedDeltaFlow_2
                ###
                # over node with BiggestInputDeltaFlow, we want as much ingoing and outgoing red flow possible, with least possible production
                OutputRedDeltaFlow = 0
                for edge in graph.out_edges(node,keys=True):
                    edge_bus_id = self.get_bus_id_from_edge(node, edge, topo_vect)
                    if edge_bus_id == Bus_BiggestInputDeltaFlow:
                        edge_color = all_edges_color_attributes[edge]
                        edge_value = float(all_edges_xlabel_attributes[edge])
                        if (edge_color == "coral"):
                            OutputRedDeltaFlow += edge_value

                min_pos_in_or_out_flows = min(OutputRedDeltaFlow, InputRedDeltaFlow)
                injection = -self.get_prod_conso_sum(node, Bus_BiggestInputDeltaFlow, topo_vect)
                final_score = np.around(min_pos_in_or_out_flows + injection, decimals=2)
        else:
            print("||||||||||||||||||||||||||| node [{}] is not connected to a path to the constrained_edge.".format(node))
            in_negative_flows_node_1 = []
            out_negative_flows_node_1 = []
            in_negative_flows_node_2 = []
            out_negative_flows_node_2 = []

            # somme des reports négatifs entrants
            for edge in graph.in_edges(node,keys=True):
                edge_bus_id = self.get_bus_id_from_edge(node, edge, topo_vect)
                edge_flow_value = float(all_edges_xlabel_attributes[edge])
                if edge_flow_value < 0:
                    if edge_bus_id == 0: # MASK
                        in_negative_flows_node_1.append(fabs(edge_flow_value))
                    elif edge_bus_id==1:
                        in_negative_flows_node_2.append(fabs(edge_flow_value))

            # somme des reports négatifs sortants
            for edge in graph.out_edges(node,keys=True):
                edge_bus_id = self.get_bus_id_from_edge(node, edge, topo_vect)
                edge_flow_value = float(all_edges_xlabel_attributes[edge])
                if edge_flow_value < 0:
                    if edge_bus_id == 0: # MASK
                        out_negative_flows_node_1.append(fabs(edge_flow_value))
                    elif edge_bus_id==1:
                        out_negative_flows_node_2.append(fabs(edge_flow_value))

            score_1=fabs(sum(in_negative_flows_node_1)-sum(out_negative_flows_node_1))
            score_2=fabs(sum(in_negative_flows_node_2)-sum(out_negative_flows_node_2))
            final_score = np.around(min(score_1,score_2),
                decimals=2)  # + diff_sums, it is not very clear its impact
            if self.debug:
                print("in negative flows node 1 = ", in_negative_flows_node_1)
                print("out negative flows node 1 = ", out_negative_flows_node_1)
                print("in negative flows node 2 = ", in_negative_flows_node_2)
                print("out negative flows node 2 = ", out_negative_flows_node_2)
                print("Final score = ", final_score)

        #=====================================================================
        # print("SCORE   ---  "+str(final_score))
        # print('\n')
        return final_score

    def get_prod_conso_sum(self, node, interesting_bus_id, topo_vect):
        total = 0
        elements = self.simulator_data["substations_elements"][node]
        for element, bus_id in zip(elements, topo_vect):
            if bus_id == interesting_bus_id:
                if isinstance(element, Consumption):
                    total = total - element.value
                elif isinstance(element, Production):
                    total = total + element.value
        return total

    def get_bus_id_from_edge(self, node, edge, topo_vect):
        """
        Knowing that topo_vect is applied on given node, returns on which bus_id is connected a given edge

        :param node:
        :param edge:
        :param topo_vect:
        :return: an int representing the bus_id on which the edge is connected to node
        """

        # Get edge substation id extremity (the other one than edge)
        target_extremity = edge[0]
        if target_extremity == node:
            target_extremity = edge[1]

        # Get elements connected to the substation (given by "node") - iterate to find whic one corresponds to edge - return corresponding bus_id
        elements = self.simulator_data["substations_elements"][node]
        paralel_edge_count_id = edge[2]
        paralel_edge_counter = 0

        for element, bus_id in zip(elements, topo_vect):
            if isinstance(element, OriginLine):
                if element.end_substation_id == target_extremity:
                    if paralel_edge_counter == paralel_edge_count_id:
                        return bus_id
                    else:  # in that case there are parallel edges between the two nodes, so deal with that
                        paralel_edge_counter += 1
            elif isinstance(element, ExtremityLine):
                if element.start_substation_id == target_extremity:
                    if paralel_edge_counter == paralel_edge_count_id:
                        return bus_id
                    else:  # in that case there are parallel edges between the two nodes, so deal with that
                        paralel_edge_counter += 1

    def is_connected_to_cpath(self, all_edges_color_attributes, all_edges_xlabel_attributes, node, edge, isSingleNode):
        edge_color = all_edges_color_attributes[edge]
        edge_value = all_edges_xlabel_attributes[edge]
        # if there is a outgoing negative blue or black edge this means we are connected to cpath.
        # therefore change to bus ID 1
        bool = (float(edge_value) < 0 and (edge_color == "blue" or edge_color == "black") and not isSingleNode)
        if bool and self.debug:
            print("\n######################################################")
            print("Node [{}] is not connected to cpath. Twin node selected...".format(node))
            print("######################################################")
        return bool

    def sort_hubs(self, hubs):
        # creates a DATAFRAME and sort it, returns the sorted hubs
        # print("================= sort_hubs =================")
        if hubs:
            df = pd.DataFrame()
            df["hubs"] = hubs

            # now for each node in hubs, get the max abs(ingoing or outgoing) flow
            flows = []

            for node in hubs:
                flow_compute_ingoing = []
                flow_compute_outgoing = []

                # print("node = ", node)

                for i, row in self.df.iterrows():
                    if row["idx_or"] == node:
                        flow_compute_outgoing.append(fabs(row["delta_flows"]))

                    if row["idx_ex"] == node:
                        flow_compute_ingoing.append(fabs(row["delta_flows"]))

                max_result = max(sum(flow_compute_ingoing), sum(flow_compute_outgoing))
                # print("all flows = ", max_result)
                flows.append(max_result)

            df["max_flows"] = flows
            df.sort_values("max_flows", ascending=False, inplace=True)
            # print(df)

            return df

        # else:
        #   raise ValueError("There are no hubs")

    def identify_routing_buses(self):
        """Categories 1 to 4
        1. Hubs
        2. all nodes that belong to c_path
        3. On //path
        4. Over Da on c_path"""

        # ALGO
        # get all nodes from c_path, loops, //paths, {set of all those nodes}
        # for nodes in interesting_nodes:
        #   classify to category 1, 2, 3, 4.
        hubs= self.g_distribution_graph.get_hubs()
        df_sorted_hubs = self.sort_hubs(hubs)
        if df_sorted_hubs is None:
            return {}
        else:
            category1 = list(df_sorted_hubs["hubs"])

            constrained_path = self.g_distribution_graph.get_constrained_path()
            set_category2 = set(constrained_path.full_n_constrained_path()) - set(category1)

            sort_redLoopBuses = sorted(self.rankedLoopBuses.items(), key=lambda x: x[1], reverse=True)
            category3 = list(set([sort_redLoopBuses[i][0] for i in range(len(sort_redLoopBuses))]))  # set()  # @TODO
            set_category4 = set(constrained_path.n_aval()) - (set(category1) | set_category2 | set(category3))

            d = {1: category1, 2: set_category2, 3: category3, 4: set_category4}
        return d

    def rank_loop_buses(self, graph, df_initial_flows):
        # self.g => overflow graph
        all_nodes_value_attributes = nx.get_node_attributes(graph, "value")
        all_edges_color_attributes = nx.get_edge_attributes(graph, "color")  # dict[edge]
        all_edges_xlabel_attributes = nx.get_edge_attributes(graph, "label")

        Strength_Bus_dic = {}
        red_loops = self.g_distribution_graph.get_loops()
        for index, loop in red_loops.iterrows():
            # for loop in self.red_loops:
            for bus in loop.Path:
                if (bus != loop.Source) & (bus != loop.Target):
                    nNode = 1
                    # TO DO:we should know if bus is 1 or 2 nodes
                    if (nNode == 1):
                        strength_measure = 0  # it will be the product of production and in_flows
                        sumInRedDeltaFlows = 0
                        sumInFlowsNotRed = 0

                        LocalProduction = 0
                        # LocalProduction = float(all_nodes_value_attributes[node])  # it is actually the sum of injections. To get only production, need to use self.simulator_data["substations_elements"][node]

                        for element in self.simulator_data["substations_elements"][bus]:
                            if isinstance(element, Production):
                                LocalProduction += (element.value)

                        for edge in self.g.in_edges(bus,keys=True):
                            edge_deltaflow_value = float(all_edges_xlabel_attributes[edge])
                            edge_color = all_edges_color_attributes[edge]
                            if (edge_color == "coral"):
                                sumInRedDeltaFlows += edge_deltaflow_value
                            else:  # we need to retrieve the initial flow from df_initial_flows
                                source, target,idx = edge

                                otherBus = source
                                if otherBus == bus:
                                    otherBus = target

                                nodes_or = df_initial_flows["idx_or"]
                                nodes_ex = df_initial_flows["idx_ex"]
                                # indexEdge_inDf=-1

                                for i in range(len(nodes_or)):
                                    flowValue = df_initial_flows["init_flows"][i]
                                    if ((flowValue >= 0) & (nodes_or[i] == otherBus) & (nodes_ex[i] == bus)):  # we are only looking for input flows
                                        indexEdge_inDf = i
                                        sumInFlowsNotRed += np.abs(flowValue)
                                        break
                                    elif ((flowValue <= 0) & (nodes_or[i] == bus) & (nodes_ex[i] == otherBus)):
                                        indexEdge_inDf = i
                                        sumInFlowsNotRed += np.abs(flowValue)
                                        break

                        sumInFlowsNotRed += LocalProduction
                        strength_measure = sumInFlowsNotRed * sumInRedDeltaFlows
                        Strength_Bus_dic[bus] = strength_measure
        return Strength_Bus_dic

    def rank_red_loops(self):
        cut_values = []
        cut_sets = []  # contains the edges that ended up having the minimum cut_values
        g_red_DiGraph=self.to_DiGraph(self.g_distribution_graph.g_only_red_components)#self.g_only_red_components)#necessary to be able to compute minimum_cut

        red_loops = self.g_distribution_graph.get_loops()
        for i, row in red_loops.iterrows():#red_loops.iterrows():
            source = row["Source"]
            target = row["Target"]
            p = row["Path"]

            # print("=============== source: {}, target: {}".format(source, target))

            #cut_value, partition = nx.minimum_cut(self.g_only_red_components, source, target)
            cut_value, partition = nx.minimum_cut(g_red_DiGraph, source, target)
            reachable, non_reachable = partition
            # print("cut_value: {}, partition: {}".format(cut_value, partition))

            # info from doc - ‘partition’ here is a tuple with the two sets of nodes that define the minimum cut.
            # You can compute the cut set of edges that induce the minimum cut as follows:
            cutset = set()
            for u, nbrs in ((n, self.g_distribution_graph.g_only_red_components[n]) for n in reachable):
                cutset.update((u, v) for v in nbrs if v in non_reachable)
            # print("sorted(cutset) = ", sorted(cutset))

            cut_values.append(cut_value)
            cut_sets.append(list(cutset)[0])

        # print("cut_values = ", cut_values)

        red_loops["min_cut_values"] = cut_values
        red_loops["min_cut_edges"] = cut_sets
        # print("======================= cut_values added =======================")
        # print(self.red_loops)

    # create weighted digraph from MultiDiGraph
    def to_DiGraph(self,gM):
        G = nx.DiGraph()
        for u, v,idx, data in gM.edges(data=True,keys=True):
            w = data['capacity'] if 'capacity' in data else 1.0
            if G.has_edge(u, v):
                G[u][v]['capacity'] += w
            else:
                G.add_edge(u, v, capacity=w)
        return G

    def compute_meaningful_structures(self):
        self.data["constrained_path"] = self.get_constrained_path()

    def get_adjacency_matrix(self, g):
        print("adjacency matrix full = ")
        for line in nx.generate_adjlist(g):
            print(line)

    def get_loop_paths(self):
        pass

    def filter_constrained_path(self, path_to_filter):
        """This function gets rid of duplicates, tuples, arrays, and return a single clean ordered array"""
        set_constrained_path = []
        for edge in path_to_filter:
            for node in edge:
                if isinstance(node, tuple):
                    for n in node:
                        if n not in set_constrained_path:
                            set_constrained_path.append(n)
                else:
                    if node not in set_constrained_path:
                        set_constrained_path.append(node)
        return set_constrained_path

    def isAntenna(self):
        pass

    def write_g(self, g):
        """This saves file g"""
        nx.write_edgelist(self.g, "./alphaDeesp/tmp/save.graph")
        # print("file saved")
        pass

    def read_g(self):
        pass


class AlphaDeesp_warmStart(AlphaDeesp):
    def __init__(self, g, g_distribution_graph,simulator_data=None, debug=False):
        # used for postprocessing
        self.bag_of_graphs = {}
        self.debug = debug
        self.boolean_dump_data_to_file = False

        # data from Simulator Class
        self.g=g
        #Compute the overload distribution graph (constrained path, loops, hubs)
        self.g_distribution_graph=g_distribution_graph
        self.simulator_data=simulator_data


#!/usr/bin/python3
# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of ExpertOp4Grid, an expert system approach to solve flow congestions in power grids

#from make_evaluation_env import make_grid2op_evaluation_env
from make_training_env import make_grid2op_training_env
from load_evaluation_data import list_all_chronics, get_first_obs_on_chronic
from datetime import datetime

import os
from alphaDeesp.core.grid2op.Grid2opObservationLoader import Grid2opObservationLoader
from alphaDeesp.core.grid2op.Grid2opSimulation import Grid2opSimulation
from alphaDeesp.core.graphsAndPaths import OverFlowGraph,Structured_Overload_Distribution_Graph

import json

from Expert_rule_action_verification import build_overflow_graph, categorize_action_space, identify_action_type, check_rules, identify_grid2op_action_type, load_interesting_lines
from packaging.version import Version as version_packaging
from importlib.metadata import version

#: name of the powerline that are now removed (non existant)
#: but are in the environment because we need them when we will use
#: historical dataset
#Note: reimporté directement ici pour éviter les problèmes de dépendances, en effet DELETED_LINE_NAME changeait de valeur après l'éxécution de categorize_action_space dans test_overflow_graph_actions_filtered
DELETED_LINE_NAME = ['BXNE L32BXNE5', 'BXNE5L32MTAGN', 'BXNE5L32CORGO', 'BXNE5L31MTAGN']

EXOP_MIN_VERSION = version_packaging("0.2.6")
if version_packaging(version("expertop4grid")) < EXOP_MIN_VERSION:
    raise RuntimeError(f"Incompatible version found for expertOp4Grid, make sureit is >= {EXOP_MIN_VERSION}")
#################################
### TEST Section
def test_overflow_graph_construction():
    """
    This function tests the construction of the overflow graph and the identification of constrained and dispatch paths.
    """
    date = datetime(2024, 8, 28)  # we choose a date for the chronic
    timestep=1#36
    line_defaut="P.SAOL31RONCI"#"FRON5L31LOUHA"
    env_name = "env_dijon_v2_assistant"
    non_connected_reconnectable_lines = ['BOISSL61GEN.P', 'CHALOL31LOUHA', 'CRENEL71VIELM', 'CURTIL61ZCUR5',
                                         'GEN.PL73VIELM',
                                         'P.SAOL31RONCI',
                                         'PYMONL61VOUGL', 'BUGEYY715', 'CPVANY632', 'GEN.PY762', 'PYMONY632']

    #env = grid2op.make(env_name, backend=backend, n_busbar=6, param=p)
    env = make_grid2op_training_env(".", env_name)#make_grid2op_evaluation_env(".", env_name)

    # make the environment
    chronics_name = list_all_chronics(env)
    print("chronics names are:")
    print(chronics_name)

    # we get the first observation for the chronic at the desired date
    obs = get_first_obs_on_chronic(date, env)

    act_deco_defaut=env.action_space({"set_line_status": [(line_defaut, -1)]})



    obs_simu, reward, done, info=obs.simulate(act_deco_defaut,time_step=timestep)

    lines_overloaded_ids=[i for i,rho in enumerate(obs_simu.rho) if rho>=1]

    param_options_test={
        # 2 percent of the max overload flow
        "ThresholdReportOfLine": 0.2,  # 0.05,#
        # 10 percent de la surcharge max
        "ThersholdMinPowerOfLoop": 0.1,
        # If at least a loop is detected, only keep the ones with a flow  of at least 25 percent the biggest one
        "ratioToKeepLoop": 0.25,
        # Ratio percentage for reconsidering the flow direction
        "ratioToReconsiderFlowDirection": 0.75,
        # max unused lines
        "maxUnusedLines": 3,
        # number of simulated topologies node at the final simulation step
        "totalnumberofsimulatedtopos": 30,
        # number of simulated topologies per node at the final simulation step
        "numberofsimulatedtopospernode": 10
    }
    inhibit_swapped_flow_reversion=False
    do_consolidate_graph=True
    lines_non_reconnectable = []  # TO DELETE, just for test
    df_of_g, overflow_sim,g_overflow,hubs,g_distribution_graph=build_overflow_graph(obs_simu, env.action_space, env.observation_space, lines_overloaded_ids,
                                                                                    non_connected_reconnectable_lines,lines_non_reconnectable, param_options_test, timestep,do_consolidate_graph,inhibit_swapped_flow_reversion)

    ##########
    # get useful paths for action verification

    lines_constrained_path, nodes_constrained_path = g_distribution_graph.get_constrained_edges_nodes()

    lines_redispatch,list_nodes_dispatch_path = g_distribution_graph.get_dispatch_edges_nodes()

    ############
    # Pour tests
    list_nodes_constrained_path_test=['NAVILP3','CPVANP6','CPVANP3','CHALOP6','GROSNP6', '1GROSP7',
                                      'GROSNP7', 'VIELMP7', 'H.PAUP7', 'SSV.OP7', 'ZCUR5P6', 'H.PAUP6', '2H.PAP7',
                                      'COUCHP6', 'VIELMP6', '1VIELP7', 'COMMUP6', 'ZMAGNP6', 'C.REGP6', 'BEON P3', 'P.SAOP3']

    list_lines_contrained_path_test=['GROSNY761','COMMUL61VIELM', 'GROSNY771', 'COUCHL61CPVAN', 'VIELMY771', 'VIELMY763', 'GROSNY762',
                                     'H.PAUL61ZCUR5', 'VIELMY762', 'CPVANY632', 'GROSNL61ZCUR5', 'C.REGL61VIELM', 'H.PAUL71VIELM',
                                     'H.PAUY762', 'CPVANY633', 'C.REGL62VIELM', 'CHALOL62GROSN', 'CHALOL61CPVAN', 'C.REGL61ZMAGN',
                                     'COMMUL61H.PAU', 'CHALOL61GROSN', 'GROSNL71SSV.O', 'CPVANL61ZMAGN', 'COUCHL61VIELM',
                                     'VIELMY761', 'BEON L31P.SAO','BEON L31CPVAN', 'H.PAUY772', 'CPVANY631','NAVILL31P.SAO']

    list_nodes_dispatch_path_test=['1GEN.P7', 'BOISSP6', 'C.FOUP3', 'CHALOP3', 'CHALOP6', 'CIZE P6', 'CPVANP6', 'CREYSP7', 'CUISEP3',
     'FLEYRP6', 'FRON5P3', 'G.CHEP3', 'GEN.PP6', 'GEN.PP7', 'GROSNP6', 'H.PAUP7', 'IZERNP6', 'LOUHAP3', 'MACONP6','MERVAP3',
     'NAVILP3', 'PYMONP3', 'PYMONP6', 'SAISSP3', 'SSUSUP3', 'SSV.OP7', 'VIELMP7', 'VOUGLP3', 'VOUGLP6','ZJOUXP6']

    list_lines_redispatch_path_test=['GEN.PY761', 'GEN.PL61IZERN', 'GEN.PL61VOUGL', 'GEN.PL73VIELM', 'GEN.PY771', 'GEN.PY762', 'GROSNL61MACON',
     'GEN.PL71VIELM', 'GEN.PL72VIELM','H.PAUL71SSV.O', 'BOISSL61GEN.P', 'CHALOL31LOUHA', 'CHALOY633', 'CHALOY631', 'CHALOY632',
     'CIZE L61FLEYR', 'CPVANL61PYMON', 'CREYSL71GEN.P', 'CREYSL72GEN.P','CUISEL31G.CHE', 'FLEYRL61VOUGL', 'FRON5L31LOUHA', 'FRON5L31G.CHE', 'CIZE L61IZERN', 'LOUHAL31SSUSU',
     'MACONL61ZJOUX', 'C.FOUL31MERVA', 'LOUHAL31PYMON', 'PYMONL61VOUGL','PYMONY632', 'PYMONL31SAISS', 'MERVAL31SSUSU',
     'CREYSL71SSV.O', 'CREYSL72SSV.O', 'CUISEL31VOUGL', 'SAISSL31VOUGL','VOUGLY632', 'VOUGLY631', 'BOISSL61ZJOUX','C.FOUL31NAVIL']

    list_hubs_test=[ 'VIELMP7', 'H.PAUP7', 'SSV.OP7','NAVILP3']#[ 'CPVANP6', 'CHALOP6', 'GROSNP6', 'VIELMP7', 'H.PAUP7', 'SSV.OP7','NAVILP3']#'P.SAOP3',

    assert(set(list_nodes_constrained_path_test)==set(nodes_constrained_path))
    assert (set(list_lines_contrained_path_test) == set(lines_constrained_path))
    assert (set(list_nodes_dispatch_path_test) == set(list_nodes_dispatch_path))
    assert (set(list_lines_redispatch_path_test) == set(lines_redispatch))
    assert (set(list_hubs_test) == set(hubs))

def test_overflow_graph_actions_filtered(check_with_action_description=True):
    """
    This function tests the filtering of actions based on the overflow graph and the identification of constrained and dispatch paths.
    It verifies that the actions are correctly categorized into filtered and unfiltered actions based on the expert rules.

    The test involves the following steps:
    1. Setting up the environment and loading the necessary data.
    2. Simulating the environment to get the initial observation.
    3. Identifying overloaded lines and building the overflow graph.
    4. Extracting constrained and dispatch paths from the overflow graph.
    5. Categorizing the actions based on the expert rules.
    6. Asserting that the number of actions and their categorization match the expected values.

    The test ensures that the expert rules are correctly applied to filter out inappropriate actions.
    """
    date = datetime(2024, 8, 28)  # we choose a date for the chronic
    timestep = 1  # 36
    line_defaut = "P.SAOL31RONCI"  # "FRON5L31LOUHA"
    env_folder = "./"
    env_name = "env_dijon_v2_assistant"

    action_space_folder = "action_space"
    file_action_space_desc = "actions_repas_most_frequent_topologies_revised.json"
    file_path = os.path.join(action_space_folder, file_action_space_desc)

    # Load actions
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"The file {file_path} does not exist.")

    with open(file_path, 'r') as file:
        dict_action = json.load(file)

    # Make the environment
    env = make_grid2op_training_env(env_folder, env_name)

    chronics_name = list_all_chronics(env)
    print("chronics names are:")
    print(chronics_name)

    # Get the first observation for the chronic at the desired date
    obs = get_first_obs_on_chronic(date, env)

    # read non reconnectable lines
    path_chronic = [path for path in env.chronics_handler.real_data.subpaths if date.strftime('%Y%m%d') in path][0]
    lines_non_reconnectable = list(load_interesting_lines(path=path_chronic,file_name="non_reconnectable_lines.csv"))
    lines_should_not_reco_2024_and_beyond =DELETED_LINE_NAME


    # simulate contingency tp detect overloads
    act_deco_defaut = env.action_space({"set_line_status": [(line_defaut, -1)]})

    obs_simu, reward, done, info = obs.simulate(act_deco_defaut, time_step=timestep)

    non_connected_reconnectable_lines = [l_name for i,l_name in enumerate(env.name_line)
                                         if l_name not in lines_non_reconnectable+lines_should_not_reco_2024_and_beyond and not obs_simu.line_status[i]]
    param_options_test = {
        # 2 percent of the max overload flow
        "ThresholdReportOfLine": 0.2,  # 0.05,#
        # 10 percent de la surcharge max
        "ThersholdMinPowerOfLoop": 0.1,
        # If at least a loop is detected, only keep the ones with a flow  of at least 25 percent the biggest one
        "ratioToKeepLoop": 0.25,
        # Ratio percentage for reconsidering the flow direction
        "ratioToReconsiderFlowDirection": 0.75,
        # max unused lines
        "maxUnusedLines": 3,
        # number of simulated topologies node at the final simulation step
        "totalnumberofsimulatedtopos": 30,
        # number of simulated topologies per node at the final simulation step
        "numberofsimulatedtopospernode": 10
    }

    inhibit_swapped_flow_reversion = True  # Cancel the swapped edge direction for swapped flows (possibly not needed anymore given the new consolidate graph functions)
    lines_overloaded_ids = [i for i, rho in enumerate(obs_simu.rho) if rho >= 1]
    do_consolidate_graph = True
    lines_non_reconnectable=[]
    df_of_g, overflow_sim, g_overflow, hubs, g_distribution_graph = build_overflow_graph(obs_simu, env.action_space, env.observation_space,
                                                            lines_overloaded_ids, non_connected_reconnectable_lines,lines_non_reconnectable,
                                                            param_options_test, timestep,do_consolidate_graph, inhibit_swapped_flow_reversion)

    ##########
    # Get useful paths for action verification
    lines_constrained_path, nodes_constrained_path = g_distribution_graph.get_constrained_edges_nodes()

    lines_dispatch, nodes_dispatch_path = g_distribution_graph.get_dispatch_edges_nodes()

    #########
    # Check rules for each action
    lines_reco_maintenance=[]
    actions_to_filter, actions_unfiltered = categorize_action_space(dict_action, hubs, lines_constrained_path, nodes_constrained_path, lines_dispatch,
                            nodes_dispatch_path, obs, timestep, line_defaut, env.action_space, lines_overloaded_ids,lines_reco_maintenance,by_description=check_with_action_description)

    n_actions = len(dict_action.keys())
    n_actions_filtered = len(actions_to_filter.keys())
    n_actions_unfiltered = len(actions_unfiltered.keys())
    n_actions_badly_filtered = len([id for id, act_filter_content in actions_to_filter.items() if act_filter_content["is_rho_reduction"]])

    # Could also directly compare to saved dictionaries "actions_to_filter_expert_rules.json" and "actions_unfiltered_expert_rules.json"
    assert(n_actions == 102)
    assert(n_actions_filtered == 58)
    assert(n_actions_unfiltered == n_actions - n_actions_filtered)
    assert(n_actions_badly_filtered == 1)  # Opening OC 'MAGNY3TR633 DJ_OC' in the substation 'MAGNYP3'. This action is filtered because of the significant delta flow threshold.
    # If "ThresholdReportOfLine" is reduced from 0.2 to 0.05, then this action is not filtered anymore, and everything works as expected

def test_action_type_open_coupling():
    actions_desc={
        "description_unitaire": " Ouverture OC 'VOUGL6COUPL DJ_OC' dans le poste 'VOUGLP6'",
        "content": {
            "set_bus": {
                "lines_or_id": {
                    "VOUGLY612": 1,
                    "VOUGLY631": 1,
                    "VOUGLY632": 2
                },
                "lines_ex_id": {
                    "FLEYRL61VOUGL": 2,
                    "GEN.PL61VOUGL": 1,
                    "PYMONL61VOUGL": 2
                },
                "loads_id": {},
                "generators_id": {},
                "shunts_id": {}
            }
        },
        "VoltageLevelId": "VOUGLP6"
    }

    action_type=identify_action_type(actions_desc, by_description=True)
    assert(action_type=="open_coupling")

def test_action_type_open_line():
    actions_desc={
        "description_unitaire": "Ouverture OC 'PYMON3TR632 DJ_OC' dans le poste 'PYMONP3'",
        "content": {
            "set_bus": {
                "lines_or_id": {},
                "lines_ex_id": {
                    "PYMONY632": -1
                },
                "loads_id": {},
                "generators_id": {},
                "shunts_id": {}
            }
        },
        "VoltageLevelId": "PYMONP3"
    }

    action_type=identify_action_type(actions_desc, by_description=True)
    assert(action_type=="open_line")

def test_action_type_open_line_load():
    actions_desc={
        "description_unitaire": " Ouverture OC 'GEN.P6CHAV6.1 DJ_OC' dans le poste 'GEN.PP6'\n- Ouverture OC 'GEN.P6AT762 DJ_OC' dans le poste 'GEN.PP6'",
        "content": {
            "set_bus": {
                "lines_or_id": {},
                "lines_ex_id": {
                    "GEN.PY762": -1
                },
                "loads_id": {
                    "CHAV6L61GEN.P":-1
                },
                "generators_id": {},
                "shunts_id": {}
            }
        },
        "VoltageLevelId": "GEN.PP6"
    }

    action_type=identify_action_type(actions_desc, by_description=True)
    assert(action_type=="open_line_load")

def test_action_type_close_line():
    actions_desc={
        "description_unitaire": "Fermeture OC 'PYMON6CPVAN.1 DJ_OC' dans le poste 'PYMONP6'(reconnection sur noeuds 1 aux 2 postes extremites)",
        "content": {
            "set_bus": {
                "lines_or_id": {
                    "CPVANL61PYMON": 1
                },
                "lines_ex_id": {
                    "CPVANL61PYMON": 1
                },
                "loads_id": {},
                "generators_id": {},
                "shunts_id": {}
            }
        },
        "VoltageLevelId": "PYMONP6"
    }

    action_type=identify_action_type(actions_desc, by_description=True)
    assert(action_type=="close_line")

def test_action_type_close_coupling():
    actions_desc = {
        "description_unitaire": "Fermeture OC 'CPVAN3COUPL DJ_OC' dans le poste 'CPVANP3'",
        "content": {
            "set_bus": {
                "lines_or_id": {
                    "CPVANL31RIBAU": 1
                },
                "lines_ex_id": {
                    "BEON L31CPVAN": 1,
                    "CPVANY631": 1,
                    "CPVANY632": 1,
                    "CPVANY633": 1
                },
                "loads_id": {
                    "ARBOIL31CPVAN": 1,
                    "BREVAL31CPVAN": 1,
                    "CPDIVL32CPVAN": 1,
                    "CPVANL31MESNA": 1,
                    "CPVANL31ZBRE6": 1,
                    "CPVAN3TR312": 1,
                    "CPVAN3TR311": 1
                },
                "shunts_id": {},
                "generators_id": {}
            }
        },
        "VoltageLevelId": "CPVANP3"
    }

    action_type = identify_action_type(actions_desc, by_description=True)
    assert (action_type == "close_coupling")

def test_action_types():
    test_action_type_close_line()
    test_action_type_open_line()
    test_action_type_open_coupling()
    test_action_type_close_coupling()
    test_action_type_open_line_load()

def test_grid2op_action_type_open_line_load(action_space):
    actions_desc={
            "set_bus": {
                "lines_or_id": {},
                "lines_ex_id": {
                    "GEN.PY762": -1
                },
                "loads_id": {
                    "CHAV6L61GEN.P":-1
                },
                "generators_id": {},
                "shunts_id": {}
            }
    }

    grid2op_action = action_space(actions_desc)
    action_type = identify_grid2op_action_type(grid2op_action)
    assert(action_type=="open_line_load")

def test_grid2op_action_type_close_line(action_space):
    actions_desc={
            "set_bus": {
                "lines_or_id": {
                    "CPVANL61PYMON": 1
                },
                "lines_ex_id": {
                    "CPVANL61PYMON": 1
                },
                "loads_id": {},
                "generators_id": {},
                "shunts_id": {}
            }
    }

    grid2op_action = action_space(actions_desc)
    action_type = identify_grid2op_action_type(grid2op_action)
    assert(action_type=="close_line")

def test_grid2op_action_type_open_line(action_space):
    actions_desc={
            "set_bus": {
                "lines_or_id": {},
                "lines_ex_id": {
                    "PYMONY632": -1
                },
                "loads_id": {},
                "generators_id": {},
                "shunts_id": {}
            }
    }

    grid2op_action = action_space(actions_desc)
    action_type = identify_grid2op_action_type(grid2op_action)
    assert(action_type=="open_line")


def test_grid2op_action_type_open_coupling(action_space):
    actions_desc={
            "set_bus": {
                "lines_or_id": {
                    "VOUGLY612": 1,
                    "VOUGLY631": 1,
                    "VOUGLY632": 2
                },
                "lines_ex_id": {
                    "FLEYRL61VOUGL": 2,
                    "GEN.PL61VOUGL": 1,
                    "PYMONL61VOUGL": 2
                },
                "loads_id": {},
                "generators_id": {},
                "shunts_id": {}
            }
    }

    grid2op_action = action_space(actions_desc)
    action_type = identify_grid2op_action_type(grid2op_action)
    assert(action_type=="open_coupling")


def test_grid2op_action_type_close_coupling(action_space):
    actions_desc={'set_bus': {'lines_or_id': {'CPVANL31RIBAU': 1},
          'lines_ex_id': {'BEON L31CPVAN': 1,
           'CPVANY631': 1,
           'CPVANY632': 1,
           'CPVANY633': 1},
          'loads_id': {'ARBOIL31CPVAN': 1,
           'BREVAL31CPVAN': 1,
           'CPDIVL32CPVAN': 1,
           'CPVANL31MESNA': 1,
           'CPVANL31ZBRE6': 1,
           'CPVAN3TR312': 1,
           'CPVAN3TR311': 1},
          'shunts_id': {},
          'generators_id': {}}}

    grid2op_action = action_space(actions_desc)
    action_type = identify_grid2op_action_type(grid2op_action)

    assert (action_type == "close_coupling")

def test_grid2op_action_types():
    #load action space
    env_name = "env_dijon_v2_assistant"
    env = make_grid2op_training_env(".", env_name)  # make_grid2op_evaluation_env(".", env_name)
    action_space=env.action_space

    # run tests
    test_grid2op_action_type_open_line_load(action_space)
    test_grid2op_action_type_close_line(action_space)
    test_grid2op_action_type_open_line(action_space)
    test_grid2op_action_type_close_coupling(action_space)
    test_grid2op_action_type_open_coupling(action_space)

def test_no_broken_rule_multi_node_dispatch_path():
    action_type="open_coupling"
    localization="dispatch_path"
    subs_topology=[[1,1,2,2,2,1]]#topology already in multi nodes
    do_filter_action, broken_rule=check_rules(action_type, localization, subs_topology)

    assert(do_filter_action==False)
    assert (broken_rule is None)

def test_broken_rule_open_line_dispatch_path():
    action_type="open_line"
    localization="dispatch_path"
    subs_topology=[]#topology already in multi nodes
    do_filter_action, broken_rule=check_rules(action_type, localization, subs_topology)

    assert(do_filter_action)
    assert (broken_rule=="No line disconnection on dispatch path")

def test_broken_rule_close_line_constrained_path():
    action_type="close_line"
    localization="constrained_path"
    subs_topology=[]#topology already in multi nodes
    do_filter_action, broken_rule=check_rules(action_type, localization, subs_topology)

    assert(do_filter_action)
    assert (broken_rule=="No line reconnection on constrained path")

def test_broken_rule_close_coupling_constrained_path():
    action_type="close_coupling"
    localization="constrained_path"
    subs_topology=[]#topology already in multi nodes
    do_filter_action, broken_rule=check_rules(action_type, localization, subs_topology)

    assert(do_filter_action)
    assert (broken_rule=="No node merging on constrained path")

def test_broken_rule_open_coupling_dispatch_path():
    action_type="open_coupling"
    localization="dispatch_path"
    subs_topology=[[1,1,1,1]]#topology in one node
    do_filter_action, broken_rule=check_rules(action_type, localization, subs_topology)

    assert(do_filter_action)
    assert (broken_rule=="No node splitting on dispatch path")

def test_load_action_no_filter():
    action_type = "open_line_load"
    localization = ""
    subs_topology = []

    do_filter_action, broken_rule = check_rules(action_type, localization, subs_topology)

    assert(do_filter_action==False)
    assert (broken_rule is None)

def test_rules():
    test_no_broken_rule_multi_node_dispatch_path()
    test_broken_rule_open_line_dispatch_path()
    test_broken_rule_close_line_constrained_path()
    test_broken_rule_close_coupling_constrained_path()
    test_broken_rule_open_coupling_dispatch_path()
    test_load_action_no_filter()

if __name__ == "__main__":

    print("STARTING TESTS")
    test_overflow_graph_actions_filtered()
    test_overflow_graph_actions_filtered(check_with_action_description=False)
    test_grid2op_action_types()
    test_overflow_graph_construction()
    test_action_types()
    test_rules()
    print("ENDING TESTS")


#!/usr/bin/python3
# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of ExpertOp4Grid, an expert system approach to solve flow congestions in power grids
__author__ = "MarcM, NMegel, mjothy"

import os
import argparse
import configparser

from alphaDeesp.core.printer import shell_print_project_header
from alphaDeesp.expert_operator import expert_operator

def main():
    # ###############################################################################################################
    # Read parameters provided by manual mode (Shell - config.ini - param_folder for the grid)
    shell_print_project_header()

    parser = argparse.ArgumentParser(description="ExpertOp4Grid")
    parser.add_argument("-d", "--debug", type=int,
                        help="If 1, prints additional information for debugging purposes. If 0, doesn't print any info", default = 0)
    parser.add_argument("-s", "--snapshot", type=int,
                        help="If 1, displays the main overflow graph at step i, ie, delta_flows_graph, diff between "
                             "flows before and after cutting the constrained line. If 0, doesn't display the graphs", default = 0)
    # nargs '+' == 1 or more.
    # nargs '*' == 0 or more.
    # nargs '?' == 0 or 1.
    # ltc for: Lines To Cut
    parser.add_argument("-l", "--ltc", nargs="+", type=int,
                        help="List of integers representing the lines to cut", default = [9])
    parser.add_argument("-t", "--timestep", type=int,
                        help="ID of the timestep to use, starting from 0. Default is 0, i.e. the first time step will be considered", default = 0)
    parser.add_argument("-c", "--chronicscenario",
                        help="Name or id of chronic scenario to consider, as stored in chronics folder. By default, the first available chronic scenario will be chosen",
                        default=0)
    parser.add_argument("-f", "--fileconfig",
                        help="Path to .ini file that provides detailed configuration of the module. If None, a default config.ini is provided in package",
                        default=None)
    args = parser.parse_args()

    if args.fileconfig is None:
        print("Getting default package config.ini")
        fileconfig = os.path.join(os.path.dirname(os.path.realpath(__file__)), "ressources",'config', "config.ini")
    else:
        fileconfig = args.fileconfig
        if not os.path.exists(fileconfig):
            raise FileNotFoundError("no config file at provided path : "+str(fileconfig))

    config = configparser.ConfigParser()
    config.read(fileconfig)
    print("#### PARAMETERS #####")
    for key in config["DEFAULT"]:
        print("key: {} = {}".format(key, config['DEFAULT'][key]))
    print("#### ########## #####\n")

    if args.ltc is None or len(args.ltc) != 1:
        raise ValueError("Input arg error, --ltc, for the moment, we allow cutting only one line.\n\nPlease select"
                         " one line to cut ex: python3 -m alphaDeesp.main -l 9")

    if args.snapshot > 1:
        raise ValueError("Input arg error, --snapshot, options are 0 or 1")

    if args.debug > 1:
        raise ValueError("Input arg error, --debug, options are 0 or 1")


    print("-------------------------------------")
    print(f"Working on lines: {args.ltc} ")
    print("-------------------------------------\n")


    # ###############################################################################################################
    # Use Loaders API to load simulator environment in manual mode at desired timestep
    sim = None
    args.debug = bool(args.debug)
    args.snapshot = bool(args.snapshot)


    if config["DEFAULT"]["simulatorType"] == "Pypownet":
        print("We init Pypownet Simulation")
        from alphaDeesp.core.pypownet.PypownetSimulation import PypownetSimulation
        from alphaDeesp.core.pypownet.PypownetObservationLoader import PypownetObservationLoader

        parameters_folder = config["DEFAULT"]["gridPath"]
        loader = PypownetObservationLoader(parameters_folder)
        env, obs, action_space = loader.get_observation(args.timestep)
        sim = PypownetSimulation(env, obs, action_space, param_options=config["DEFAULT"], debug=args.debug,
                                 ltc=args.ltc)
    elif config["DEFAULT"]["simulatorType"] == "Grid2OP":
        print("We init Grid2OP Simulation")
        from alphaDeesp.core.grid2op.Grid2opSimulation import Grid2opSimulation
        from alphaDeesp.core.grid2op.Grid2opObservationLoader import Grid2opObservationLoader

        try:
            parameters_folder = config["DEFAULT"]["gridPath"] # Case there is a grid path given in config.ini
        except: # Default load l2rpn_2019 in packages data
            print("Getting default package grid l2rpn_2019")
            parameters_folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), "ressources",
                                                                                            'parameters', "l2rpn_2019")
            config["DEFAULT"]["gridPath"] = parameters_folder
        try:
            difficulty = str(config["DEFAULT"]["grid2opDifficulty"])
        except:
            print("Default difficulty level has been set to None")
            difficulty = None
        loader = Grid2opObservationLoader(parameters_folder, difficulty = difficulty)
        env, obs, action_space = loader.get_observation(chronic_scenario= args.chronicscenario, timestep=args.timestep)
        observation_space = env.observation_space


        # Lok for scenario Name if none has been given (first one by default)
        if args.chronicscenario is None:
            args.chronicscenario = loader.search_chronic_name_from_num(0)

        # Create plot folders locally
        if args.snapshot:
            try:
                plot_base_folder = config["DEFAULT"]["outputPath"] # Case there is a grid path given in config.ini
            except: # Default load l2rpn_2019 in packages data
                print("No outputPath in config.ini: generating outputs in current folder")
                plot_base_folder = "output"
            plot_folder = generate_plot_folders(plot_base_folder, args.ltc,args.chronicscenario,args.timestep, config)
        else:
            plot_folder = None

        sim = Grid2opSimulation(obs, action_space, observation_space, param_options=config["DEFAULT"], debug=args.debug,
                                 ltc=args.ltc, plot=args.snapshot, plot_folder = plot_folder)

    elif config["DEFAULT"]["simulatorType"] == "RTE":
        print("We init RTE Simulation")
        # sim = RTESimulation()
        return
    else:
        print("Error simulator Type in config.ini not recognized...")

    # ###############################################################################################################
    # Call agent mode with possible plot and debug fonctionalities
    ranked_combinations, expert_system_results, action = expert_operator(sim, plot=args.snapshot)

    #return ranked_combinations#, expert_system_results, action #no mean to return some value in a main function of a package that will be called in a cli


def generate_plot_folders(plot_folder, ltc, chronicscenario,timestep, config):
    os.makedirs(plot_folder, exist_ok=True)
    gridName = os.path.basename(os.path.normpath(config['DEFAULT']['gridPath']))    
    plot_folder = os.path.join(plot_folder, gridName)
    os.makedirs(plot_folder, exist_ok=True)
    lineName = 'linetocut_' + str(ltc[0])
    plot_folder = os.path.join(plot_folder, lineName)
    os.makedirs(plot_folder, exist_ok=True)
    scenarioName = 'Scenario_' + str(chronicscenario)
    plot_folder = os.path.join(plot_folder, scenarioName)
    os.makedirs(plot_folder, exist_ok=True)
    timestepName = 'Timestep_' + str(timestep)
    plot_folder = os.path.join(plot_folder, timestepName)
    os.makedirs(plot_folder, exist_ok=True)
    return plot_folder

if __name__ == "__main__":
    main()



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of ExpertOp4Grid, an expert system approach to solve flow congestions in power grids

"""This file contains all the possible displays for graph from NetworkX"""

import os
import pprint
import datetime
import networkx as nx
import pydot

import subprocess

from pathlib import Path


class Printer:
    save_id_number = 0

    def __init__(self, output_path = None):
        if output_path is None:
            self.default_output_path = "alphaDeesp/ressources/output"
        else:
            self.default_output_path = output_path
        self.base_output_path = os.path.join(self.default_output_path,"Base graph")
        os.makedirs(self.base_output_path, exist_ok=True)
        self.results_output_path = os.path.join(self.default_output_path,"Result graph")
        os.makedirs(self.results_output_path, exist_ok=True)
        print("self.default output path = ", self.default_output_path)

    def plot_graphviz(self, g, custom_layout=None,rescale_factor=None,allow_overlap=True,fontsize=None,node_thickness=3, axial_symetry=False, save=False, name=None):
        "filenames are pathlib.Paths objects"

        dic_pos_attributes = {}
        if custom_layout is not None:
            if rescale_factor is not None:
                custom_layout=[(e[0]/rescale_factor,e[1]/rescale_factor) for e in custom_layout]
            assert isinstance(custom_layout, list) is True
            # we create a dictionary to add a position attribute to the nodes
            ii = 0
            n_layout=len(custom_layout)
            for i, node in enumerate(g.nodes):#(custom_layout):
                # i += 1
                # print(f"i:{i} value:{value}")
                value=custom_layout[i%n_layout]
                if i < len(custom_layout):
                    dic_pos_attributes[node] = {"pos": (str(value[0]) + ", " + str(value[1]) + "!")}
                else:
                    node = int("666" + str(ii))
                    dic_pos_attributes[node] = {"pos": (str(value[0]) + ", " + str(value[1]) + "!")}
                    ii += 1

            # we update the graph with some specific position
            # see here "node_attributes" for more attributes you can update in the drawing https://github.com/pydot/pydot/blob/a892962a2db1a71f5e0aa83cfa734720ce2bb077/src/pydot/core.py#L61
            nx.set_node_attributes(g, dic_pos_attributes)

        if fontsize is not None:
            nx.set_node_attributes(g, fontsize,"fontsize")
            nx.set_node_attributes(g, 0, "margin")

        nx.set_node_attributes(g, node_thickness, "penwidth")#size of nodes thickness

            #nx.set_edge_attributes(g,overlap_margin,"len")

        graph = nx.drawing.nx_pydot.to_pydot(g)

        if custom_layout is not None:
            prog=["neato","-n","-x"] #-n to minimize overlap, -x to prune isolated components
            if not allow_overlap:
                graph.set_overlap(False)
            output_graphviz_svg = graph.create_svg(prog=prog)
        else:
            output_graphviz_svg = graph.create_svg(prog="dot")#(prog="dot")

        return output_graphviz_svg



    def display_geo(self, g, custom_layout=None,rescale_factor=None,fontsize=None, node_thickness=3, axial_symetry=False, save=False, name=None):
        """This function displays the graph g in a "geographical" way"""

        "filenames are pathlib.Paths objects"
        type_ = "results"
        if name in ["g_pow", "g_overflow_print", "g_pow_prime"]:
            type_ = "base"
        filename_dot, filename_pdf = self.create_namefile("geo", name=name, type = type_)

        dic_pos_attributes = {}
        if custom_layout is not None:
            if rescale_factor is not None:
                custom_layout=[(e[0]/rescale_factor,e[1]/rescale_factor) for e in custom_layout]
            assert isinstance(custom_layout, list) is True
            # we create a dictionary to add a position attribute to the nodes
            ii = 0
            n_layout=len(custom_layout)
            for i, node in enumerate(g.nodes):#(custom_layout):
                # i += 1
                # print(f"i:{i} value:{value}")
                value=custom_layout[i%n_layout]
                if i < len(custom_layout):
                    dic_pos_attributes[node] = {"pos": (str(value[0]) + ", " + str(value[1]) + "!")}
                else:
                    node = int("666" + str(ii))
                    dic_pos_attributes[node] = {"pos": (str(value[0]) + ", " + str(value[1]) + "!")}
                    ii += 1

            # we update the graph with some specific position
            # see here "node_attributes" for more attributes you can update in the drawing https://github.com/pydot/pydot/blob/a892962a2db1a71f5e0aa83cfa734720ce2bb077/src/pydot/core.py#L61
            nx.set_node_attributes(g, dic_pos_attributes)

        if fontsize is not None:
            nx.set_node_attributes(g, fontsize,"fontsize")
            nx.set_node_attributes(g, 0, "margin")

        nx.set_node_attributes(g, node_thickness, "penwidth")#size of nodes thickness

            #nx.set_edge_attributes(g,overlap_margin,"len")

        nx.drawing.nx_pydot.write_dot(g, filename_dot)

        if custom_layout is None:

            cmd_line = 'dot -Tpdf "' + str(filename_dot) + '" -o "' + str(filename_pdf) + '"'#'neato -Tpdf "' + str(filename_dot) + '" -o "' + str(filename_pdf) + '"'
        else:
            cmd_line = 'neato -n -Tpdf "' + str(filename_dot) + '" -o "' + str(filename_pdf) + '"'
        print("we print the cmd line = ", cmd_line)

        assert execute_command(cmd_line)


        #cmd_line = f"evince {str(filename_pdf)} &"
        #os.system(cmd_line)
#
        #os.system(cmd_line)
        #os.system("evince " + str(filename_pdf) + " &")

        # if save is False:
        #     assert(alphadeesp.execute_command(f"rm {filename_dot}"))
        #     assert(alphadeesp.execute_command(f"rm {filename_pdf}"))

        # if save is False:
        #     os.system("rm " + filename_dot)
        #     os.system("rm " + filename_pdf)

    def display_elec(self, g, save=False):
        pass

    def create_namefile(self, display_type, name=None, type = "results"):
        """return dot and pdf filenames"""
        # filename_dot = "graph_result_" + display_type + "_" + current_date + ".dot"
        # filename_pdf = "graph_result_" + display_type + "_" + current_date + ".pdf"
        current_date_no_filter = datetime.datetime.now()
        current_date = current_date_no_filter.strftime("%Y-%m-%d_%H-%M")
        if type == 'results':
            current_date += "_" + str(Printer.save_id_number) + "_"
            Printer.save_id_number += 1

        if name is None:
            name = ""
        print("name = ", name)
        filename_dot = name + "_" + display_type + "_" + current_date + ".dot"
        filename_pdf = name + "_" + display_type + "_" + current_date + ".pdf"

        if type == "results":
            output_path = self.results_output_path
        elif type == "base":
            output_path = self.base_output_path
        else:
            output_path = self.default_output_path
        hard_filename_dot = os.path.join(output_path,filename_dot)
        # hard_filename_dot = filename_dot
        hard_filename_pdf = os.path.join(output_path,filename_pdf)

        print("============================= FUNCTION create_namefile =============================")
        print("hard_filename = ", hard_filename_pdf)

        return hard_filename_dot, hard_filename_pdf


def shell_print_project_header():
    os.system("cat ./print_header.txt")

########################################################################################################################
# ######################################### EXTERNAL COMMANDS ##########################################################
########################################################################################################################


def execute_command(command: str):
    """
    This function executes a command on the local machine, and fill self.output and self.error with results of
    command.
    @return True if command went through
    """

    # # print("command = ", command)
    sub_p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, stderr = sub_p.communicate()
    exit_code = sub_p.returncode
    # pid = sub_p.pid

    output = stdout.decode()
    error = stderr.decode()

    # print("--------------------\n output is:", output)
    # print("--------------------\n stderr is:", error)
    # print("--------------------\n exit code is:", exit_code)
    # # print("--------------------\n pid is:", pid)

    if not error:
        # string error is empty
        return True
    else:
        # string error is full
        # # print(f"Error {error}")
        return False



import unittest
import configparser
import os, shutil
import numpy as np
from alphaDeesp.expert_operator import expert_operator
from alphaDeesp.main import generate_plot_folders
from alphaDeesp.core.grid2op.Grid2opObservationLoader import Grid2opObservationLoader
from alphaDeesp.core.grid2op.Grid2opSimulation import Grid2opSimulation

def build_sim():
    config = configparser.ConfigParser()
    config.read("./alphaDeesp/tests/resources_for_tests_grid2op/config_for_tests.ini")
    param_folder = "./alphaDeesp/tests/resources_for_tests_grid2op/l2rpn_2019_ltc_9"

    loader = Grid2opObservationLoader(param_folder)
    env, obs, action_space = loader.get_observation()
    observation_space = env.observation_space

    ltc=[9]
    chronicscenario=0
    timestep=0

    try:
        plot_base_folder = config["DEFAULT"]["outputPath"]  # Case there is a grid path given in config.ini
    except:  # Default load l2rpn_2019 in packages data
        print("No outputPath in config.ini: generating outputs in current folder")
        plot_base_folder = "output"
    plot_folder = generate_plot_folders(plot_base_folder, ltc,chronicscenario,timestep, config)

    sim = Grid2opSimulation(obs, action_space, observation_space, param_options=config["DEFAULT"], debug=False,
                                 ltc=[9],plot=True, plot_folder = plot_folder)
    return sim,env

def get_folder_empty(folder):
    for filename in os.listdir(folder):
        file_path = os.path.join(folder, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))


class TestExpertOp(unittest.TestCase):

    def test_expertOp(self):
        # test results: ranked_combinations, expert_system_results, actions
        sim,env=build_sim()

        graphs_folder=os.path.join(sim.plot_folder,"Base graph")
        plot_results_folder=os.path.join(sim.plot_folder,"Result graph")
        get_folder_empty(graphs_folder)
        get_folder_empty(plot_results_folder)
        expert_operator(sim, plot=True)#we test if plots are properly generated

        print("check if files generated")
        filenames=os.listdir(graphs_folder)

        assert len(filenames)==4
        assert np.any(["g_overflow" in filename for filename in filenames])
        assert np.any([".dot" in filename for filename in filenames])
        assert np.any(["g_pow" in filename for filename in filenames])
        assert np.any(["g_pow_prime" in filename for filename in filenames])

        filenames = os.listdir(plot_results_folder)
        assert len(filenames) == 42




import unittest
import subprocess

import numpy as np
import pandas as pd
import pathlib

from alphaDeesp import main
#from chronix2grid import constants as cst
#import chronix2grid.generation.generation_utils as gu


class TestCli(unittest.TestCase):

    def test_cli(self):
        cmd = ['expertop4grid',
            '-l', str(9),
            '-s',str(0),
            '-c',str(0),
            '-t',str(0)]

        rv = subprocess.run(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

        self.assertEqual(rv.returncode, 0)
